/////////////////////////////////////////////////////////////////////////////
// Implementation for a MFC class encapsulation of the synchronous sockets

/*
 * Copyright (c) 2003 - 2005 by Daniel Weber (daniel@astolat.de)
 * All rights reserved.
 *
 * This is a SSL/TLS-implementation for Windows-Sockets in C++ written by
 * Daniel Weber (daniel@astolat.de) based an examples provided by
 * Martin Ziacek (martin.ziacek@blueyonder.co.uk).
 *
 * You are allowed to use, modify and redistribute this source-code for
 * commercial and non-commercial software as long as the following
 * conditions are met:
 * - Copyright remains bei Daniel Weber and redistributions of the source-
 *   code must retain this copyright-notice, list of conditions and
 *   disclaimer.
 * - Redistributions in binary form must mention Daniel Weber as author
 *   of parts of the SSL/TLS-implementations at program startup and/or
 *   in the documentation.
 *
 * This source-code is provided by Daniel Weber "as is" and any express or
 * implied warranties, including but not limited to the implied warranties
 * of merchantability and fitness for a particular purpose are disclaimed.
 * The authors or contributors are not liable for any damages arising out
 * of the use of this source-code, even if advised of the possibility of
 * such damage.
 */

#include "stdafx.h"

#include "SyncSocket.h"

IMPLEMENT_DYNAMIC(CSyncSocket, CObject)

/////////////////////////////////////////////////////////////////////////////
// CSyncSocket Construction

CSyncSocket::CSyncSocket()
{
	m_hSocket = INVALID_SOCKET;
}

CSyncSocket::CSyncSocket( const CSyncSocket& rSrc )
{
	m_hSocket = rSrc.m_hSocket;
}

void CSyncSocket::operator=( const CSyncSocket& rSrc )
{
	m_hSocket = rSrc.m_hSocket;
}

CSyncSocket::~CSyncSocket()
{
	Close();
}

BOOL CSyncSocket::Create(UINT nSocketPort, int nSocketType, LPCTSTR lpszSocketAddress)
{
	if( Socket( nSocketType ) )
	{
		return Bind( nSocketPort, lpszSocketAddress );
	}
	else
	{
		return FALSE;
	}
}

/////////////////////////////////////////////////////////////////////////////
// CSyncSocket Attributes

CSyncSocket::operator SOCKET() const
{
	ASSERT( m_hSocket != INVALID_SOCKET );

	return m_hSocket;
}

BOOL CSyncSocket::GetPeerName(CString& rPeerAddress, UINT& rPeerPort)
{
	SOCKADDR_IN sockAddr;
	::memset( &sockAddr, 0, sizeof( sockAddr ) );

	int nSockAddrLen = sizeof( sockAddr );
	if( !GetPeerName( static_cast<SOCKADDR*>( &sockAddr ), &nSockAddrLen ) )
	{
		return FALSE;
	}

	rPeerPort = ::ntohs( sockAddr.sin_port );
	rPeerAddress = ::inet_ntoa( sockAddr.sin_addr );

	return TRUE;
}

BOOL CSyncSocket::GetPeerName(SOCKADDR* lpSockAddr, int* lpSockAddrLen)
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::getpeername( m_hSocket, lpSockAddr, lpSockAddrLen ) != 0 )
	{
		return FALSE;
	}

	return TRUE;
}

BOOL CSyncSocket::GetSockName(CString& rSocketAddress, UINT& rSocketPort)
{
	SOCKADDR_IN sockAddr;
	::memset( &sockAddr, 0, sizeof( sockAddr ) );

	int nSockAddrLen = sizeof( sockAddr );
	if( !GetSockName( static_cast<SOCKADDR*>( &sockAddr ), &nSockAddrLen ) )
	{
		return FALSE;
	}

	rSocketPort = ::ntohs( sockAddr.sin_port );
	rSocketAddress = ::inet_ntoa( sockAddr.sin_addr );

	return TRUE;
}

BOOL CSyncSocket::GetSockName(SOCKADDR* lpSockAddr, int* lpSockAddrLen)
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::getsockname( m_hSocket, lpSockAddr, lpSockAddrLen ) != 0 )
	{
		return FALSE;
	}

	return TRUE;
}

BOOL CSyncSocket::SetSockOpt(int nOptionName, const void* lpOptionValue, int nOptionLen, int nLevel)
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::setsockopt( m_hSocket, nLevel, nOptionName, static_cast<LPCSTR>( lpOptionValue ), nOptionLen ) != 0 )
	{
		return FALSE;
	}

	return TRUE;
}

BOOL CSyncSocket::GetSockOpt(int nOptionName, void* lpOptionValue, int* lpOptionLen, int nLevel)
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::getsockopt( m_hSocket, nLevel, nOptionName, static_cast<LPSTR>( lpOptionValue ), lpOptionLen ) != 0 )
	{
		return FALSE;
	}

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CAscynSocket Operations

BOOL CSyncSocket::Accept(CSyncSocket& rConnectedSocket, SOCKADDR* lpSockAddr, int* lpSockAddrLen)
{
	SOCKET hSocket = ::accept( m_hSocket, lpSockAddr, lpSockAddrLen );
	if( hSocket == INVALID_SOCKET )
	{
		return FALSE;
	}
	
	rConnectedSocket.m_hSocket = hSocket;

	return TRUE;
}

BOOL CSyncSocket::Bind( UINT nSocketPort, LPCTSTR lpszSocketAddress )
{
	SOCKADDR_IN sa;
	::memset( &sa, 0, sizeof( sa ) );

	// For correct operation of the T2A macro, see MFC Tech Note 59
	USES_CONVERSION;

	LPSTR lpszAscii = const_cast<LPSTR>( T2A( lpszSocketAddress ) );
	sa.sin_family = AF_INET;

	if( lpszAscii == NULL || strlen( lpszAscii ) == 0 )
	{
		sa.sin_addr.s_addr = ::htonl( INADDR_ANY );
	}
	else
	{
		DWORD lResult = ::inet_addr( lpszAscii );
		if( lResult == INADDR_NONE )
		{
			::WSASetLastError( WSAEINVAL );
			return FALSE;
		}
		sa.sin_addr.s_addr = lResult;
	}

	sa.sin_port = ::htons( static_cast<u_short>( nSocketPort ) );

	return Bind( static_cast<SOCKADDR*>( &sa ), sizeof( sa ) );
}

BOOL CSyncSocket::Bind(const SOCKADDR* lpSockAddr, int nSockAddrLen)
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::bind( m_hSocket, lpSockAddr, nSockAddrLen ) != 0 )
	{
		return FALSE;
	}

	return TRUE;
}

BOOL CSyncSocket::Close()
{
	if( m_hSocket == INVALID_SOCKET )
	{
		return FALSE;
	}

	if( ::closesocket( m_hSocket ) != 0 )
	{
		return FALSE;
	}

	m_hSocket = INVALID_SOCKET;

	return TRUE;
}

BOOL CSyncSocket::Connect( LPCTSTR lpszHostAddress, UINT nHostPort )
{
	// Must have a valid host
	ASSERT( lpszHostAddress );
	ASSERT( _tcslen( lpszHostAddress ) );

	// For correct operation of the T2A macro, see MFC Tech Note 59
	USES_CONVERSION;

	SOCKADDR_IN sa;
	::memset( &sa, 0, sizeof( sa ) );

	LPSTR lpszAscii = const_cast<LPSTR>( T2A( lpszHostAddress ) );

	// Determine if the address is in dotted notation
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = inet_addr( lpszAscii );
	sa.sin_port = ::htons( static_cast<u_short>( nHostPort ) );

	// If the address is not dotted notation, then go to DNS 
	if( sa.sin_addr.s_addr == INADDR_NONE )
	{
		LPHOSTENT lphost = ::gethostbyname( lpszAscii );
		if( lphost != NULL )
		{
			sa.sin_addr.s_addr = ( reinterpret_cast<LPIN_ADDR>( lphost->h_addr ) )->s_addr;
		}
		else
		{
			::WSASetLastError( WSAEINVAL );
			return FALSE;
		}
	}

	return Connect( static_cast<SOCKADDR*>( &sa ), sizeof( sa ) );
}

BOOL CSyncSocket::Connect( const SOCKADDR* lpSockAddr, int nSockAddrLen )
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::connect( m_hSocket, lpSockAddr, nSockAddrLen ) != 0 )
	{
		return FALSE;
	}

	return TRUE;
}

BOOL CSyncSocket::IOCtl( long lCommand, DWORD* lpArgument )
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::ioctlsocket( m_hSocket, lCommand, lpArgument ) != 0 )
	{
		return FALSE;
	}

	return TRUE;
}

BOOL CSyncSocket::Listen( int nConnectionBacklog )
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::listen( m_hSocket, nConnectionBacklog ) != 0 )
	{
		return FALSE;
	}

	return TRUE;
}

int CSyncSocket::Receive( void* lpBuf, int nBufLen, int nFlags )
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	return ::recv( m_hSocket, static_cast<LPSTR>( lpBuf ), nBufLen, nFlags );
}

int CSyncSocket::ReceiveFrom( void* lpBuf, int nBufLen, CString& rSocketAddress, UINT& rSocketPort, int nFlags )
{
	SOCKADDR_IN sockAddr;
	::memset( &sockAddr, 0, sizeof( sockAddr ) );

	int nSockAddrLen = sizeof( sockAddr );
	int nResult = ReceiveFrom( lpBuf, nBufLen, static_cast<SOCKADDR*>( &sockAddr ), &nSockAddrLen, nFlags );
	if( nResult != SOCKET_ERROR )
	{
		rSocketPort = ::ntohs( sockAddr.sin_port );
		rSocketAddress = ::inet_ntoa( sockAddr.sin_addr );
	}
	return nResult;
}

int CSyncSocket::ReceiveFrom( void* lpBuf, int nBufLen, SOCKADDR* lpSockAddr, int* lpSockAddrLen, int nFlags )
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	return ::recvfrom( m_hSocket, static_cast<LPSTR>( lpBuf ), nBufLen, nFlags, lpSockAddr, lpSockAddrLen );
}

BOOL CSyncSocket::ShutDown(int nHow )
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::shutdown( m_hSocket, nHow ) != 0 )
	{
		return FALSE;
	}

	return TRUE;
}

int CSyncSocket::Send( const void* lpBuf, int nBufLen, int nFlags )
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	return ::send( m_hSocket, static_cast<LPCSTR>( lpBuf ), nBufLen, nFlags );
}

int CSyncSocket::SendTo( const void* lpBuf, int nBufLen, UINT nHostPort, LPCTSTR lpszHostAddress, int nFlags )
{
	// Must have a valid host
	ASSERT( lpszHostAddress );
	ASSERT( _tcslen( lpszHostAddress ) );

	// For correct operation of the T2A macro, see MFC Tech Note 59
	USES_CONVERSION;

	SOCKADDR_IN sockAddr;
	::memset( &sockAddr, 0, sizeof( sockAddr ) );

	LPSTR lpszAscii = T2A( const_cast<LPTSTR>( lpszHostAddress ) );

	// Determine if the address is in dotted notation
	sockAddr.sin_family = AF_INET;

	if( lpszAscii == NULL )
	{
		sockAddr.sin_addr.s_addr = ::htonl( INADDR_BROADCAST );
	}
	else
	{
		sockAddr.sin_addr.s_addr = ::inet_addr( lpszAscii );
		if( sockAddr.sin_addr.s_addr == INADDR_NONE )
		{
			LPHOSTENT lphost = ::gethostbyname( lpszAscii );
			if( lphost != NULL )
			{
				sockAddr.sin_addr.s_addr = ( reinterpret_cast<LPIN_ADDR>( lphost->h_addr ) )->s_addr;
			}
			else
			{
				::WSASetLastError( WSAEINVAL );
				return FALSE;
			}
		}
	}

	sockAddr.sin_port = ::htons( static_cast<u_short>( nHostPort ) );

	return SendTo(lpBuf, nBufLen, static_cast<SOCKADDR*>( &sockAddr ), sizeof(sockAddr), nFlags);
}

int CSyncSocket::SendTo( const void* lpBuf, int nBufLen, const SOCKADDR* lpSockAddr, int nSockAddrLen, int nFlags )
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	return ::sendto( m_hSocket, static_cast<LPCSTR>( lpBuf ), nBufLen, nFlags, lpSockAddr, nSockAddrLen );
}

/////////////////////////////////////////////////////////////////////////////
// CSyncSocket Implementation

BOOL CSyncSocket::Socket(int nSocketType, int nProtocolType, int nAddressFormat)
{
	if( m_hSocket != INVALID_SOCKET )
	{
		return FALSE;
	}

	m_hSocket = ::socket( nAddressFormat, nSocketType, nProtocolType );
	if( m_hSocket == INVALID_SOCKET )
	{
		return FALSE;
	}

	return TRUE;
}

#ifdef _DEBUG
void CSyncSocket::AssertValid() const
{
}

void CSyncSocket::Dump(CDumpContext& dc) const
{
}

#endif

BOOL CSyncSocket::IsReadible( int nTimeout ) const
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	fd_set fds;
	FD_ZERO( &fds );
	FD_SET( m_hSocket, &fds );
	timeval tvTimeout = { nTimeout, 0 };

	if( ::select( 0, &fds, NULL, NULL, ( nTimeout != -1 ? &tvTimeout : NULL ) ) != 1 )
	{
		return FALSE;
	}

	return TRUE;
}

BOOL CSyncSocket::IsWritible( int nTimeout ) const
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	fd_set fds;
	FD_ZERO( &fds );
	FD_SET( m_hSocket, &fds );
	timeval tvTimeout = { nTimeout, 0 };

	if( ::select( 0, NULL, &fds, NULL, ( nTimeout != -1 ? &tvTimeout : NULL ) ) != 1 )
	{
		return FALSE;
	}

	return TRUE;
}

CString CSyncSocket::GetLastErrorString()
{
	int nError = GetLastError();
	CString strError;
	switch( nError )
	{
	case 0:
		strError = _T("no error");
		break;
	case WSANOTINITIALISED:
		strError = _T("a successful AfxSocketInit must occur before using this API");
		break;
	case WSAENETDOWN:
		strError = _T("the Windows Sockets implementation detected that the network subsystem failed");
		break;
	case WSAEADDRINUSE:
		strError = _T("the specified address is already in use");
		break;
	case WSAEINPROGRESS:
		strError = _T("a blocking Windows Sockets call is in progress");
		break;
	case WSAEADDRNOTAVAIL:
		strError = _T("the specified address is not available from the local machine");
		break;
	case WSAESOCKTNOSUPPORT:
		strError = _T("the specified socket type is not supported in this address family");
		break;
	case WSAEAFNOSUPPORT:
		strError = _T("addresses in the specified family cannot be used with this socket");
		break;
	case WSAEPROTONOSUPPORT:
		strError = _T("the specified port is not supported");
		break;
	case WSAEPROTOTYPE:
		strError = _T("the specified port is the wrong type for this socket");
		break;
	case WSAECONNREFUSED:
		strError = _T("the attempt to connect was rejected");
		break;
	case WSAEDESTADDRREQ:
		strError = _T("a destination address is required");
		break;
	case WSAEFAULT:
		strError = _T("the nSockAddrLen argument is incorrect");
		break;
	case WSAEINVAL:
		strError = _T("the socket is not already bound to an address");
		break;
	case WSAEISCONN:
		strError = _T("the socket is already connected");
		break;
	case WSAEMFILE:
		strError = _T("no more file descriptors are available");
		break;
	case WSAENETUNREACH:
		strError = _T("the network cannot be reached from this host at this time");
		break;
	case WSAENOBUFS:
		strError = _T("no buffer space is available, the socket cannot be created");
		break;
	case WSAENOTSOCK:
		strError = _T("the descriptor is not a socket");
		break;
	case WSAETIMEDOUT:
		strError = _T("the attempt to connect timed out without establishing a connection");
		break;
	case WSAEWOULDBLOCK:
		strError = _T("the socket is marked as nonblocking and the connection cannot be completed immediately");
		break;
	case WSAEINTR:
		strError = _T("the (blocking) call was canceled using WSACancelBlockingCall");
		break;
	case WSAENOTCONN:
		strError = _T("the socket is not connected");
		break;
	case WSAEOPNOTSUPP:
		strError = _T("MSG_OOB was specified, but the socket is not of type SOCK_STREAM");
		break;
	case WSAESHUTDOWN:
		strError = _T("the socket has been shut down; it is not possible to recv on a socket after shutdown has been invoked with how set to 0 or 2");
		break;
	case WSAEMSGSIZE:
		strError = _T("the datagram was too large to fit into the specified buffer and was truncated");
		break;
	case WSAECONNABORTED:
		strError = _T("the virtual circuit was aborted due to timeout or other failure");
		break;
	case WSAECONNRESET:
		strError = _T("the virtual circuit was reset by the remote side");
		break;
	default:
		strError.Format( _T("unknown error: 0x%08X"), nError );
		break;
	}

	return strError;
}
